# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""EMVA 1288 descriptor parser
This class takes an EMVA1288 descriptor file and loads its content into a
python dictionary.

An EMVA1288 descriptor file is a file that contains the description of an
EMVA1288 test including exposure times, photon count and corresponding images

"""

import numpy as np
import os
import pprint
import logging


class ParseEmvaDescriptorFile(object):
    '''Take an image descriptor file and transform it into
    an usable directory
    '''

    def __init__(self, filename, loglevel=logging.INFO):
        # For dark the items are in the form of
        # exposure:{[fname1, fname2, ...]}
        # for bright the items are in the form of
        # exposure:{photons:[fname1, fname2,...]}, photons....}

        self.info = {'filename': None,
                     'path': None,
                     'version': None,
                     'format': {},  # bits, witdth, height
                     'temporal': {'dark': {}, 'bright': {}},
                     'spatial': {'dark': {}, 'bright': {}},
                     'camera_info': {},
                     'operation_point_info': {},
                     }

        logging.basicConfig()
        self.log = logging.getLogger('Parser')
        self.log.setLevel(loglevel)

        self._load_file(filename)
        self._fill_info()
        self.log.debug(pprint.pformat(self.info))

    def _get_images_filenames(self):
        '''
        From the current line in self._lines array
        get all the consecutive "i filename"
        if less than 2 consecutive, raise an error
        '''
        fnames = []
        while self._lines:
            line = self._lines.pop()
            l = self._split_line(line)
            if l[0] != 'i':
                # Ups, to the end of images, reappend last line that is not an
                # image line
                self._lines.append(line)
                break
            if len(l) != 2:
                raise SyntaxError('Wrong format: "%s" should be "i filename"'
                                  % line)
                break

            npath = os.path.normpath(l[1])
            path = os.path.join(*npath.split('\\'))
            fnames.append(path)

        if len(fnames) < 2:
            raise SyntaxError('Each image series, has to '
                              'have at least two images')

        return fnames

    def _get_kind(self, fnames):
        '''
        Guess what kind of data based on the number of images
        '''
        L = len(fnames)
        if L == 2:
            kind = 'temporal'
        else:
            kind = 'spatial'

        return kind

    def _add_bright_pcount(self, exposure, photons, fnames):
        '''
        For a given exposure and photon count
        add the appropiate image filenames to the self.info dict
        '''
        kind = self._get_kind(fnames)
        step = self.info[kind]['bright'].setdefault(exposure, {})

        if photons in step:
            raise SyntaxError('Only one set of bright images per photon count '
                              '%.3f' % (photons))

        step[photons] = fnames

    def _add_dark_pcount(self, exposure, fnames):
        kind = self._get_kind(fnames)

        if exposure in self.info[kind]['dark']:
            raise SyntaxError('Only one set of images per dark exposure %.3f'
                              % (exposure))
        self.info[kind]['dark'][exposure] = fnames

    def _fill_info(self):
        '''
        Walk trhought all the lines in the descriptor file
        by the first character fill self.info
        This info is what should be the descriptor
        (too late to change to this?)
        '''

        self._lines.reverse()
        while self._lines:
            line = self._lines.pop()
            l = self._split_line(line)

            if l[0] == 'v':
                self.info['version'] = l[1]
                self.log.info('Version ' + l[1])
                continue

            if l[0] == 'n':
                if len(l) != 4:
                    raise SyntaxError('Wrong format: "%s" should be "n bits '
                                      'width height"' % line)

                if self.info.get('format'):
                    raise SyntaxError('Only one "n bits width height" is '
                                      'allowed per file')

                self.info['format']['bits'] = int(l[1])
                self.info['format']['width'] = int(l[2])
                self.info['format']['height'] = int(l[3])
                continue

            if l[0] == 'b':
                if len(l) != 3:
                    raise SyntaxError('Wrong format: "%s" should be "b '
                                      'exposure photons"' % line)

                exposure = np.float(l[1].replace(',', '.'))
                photons = np.float(l[2].replace(',', '.'))
                fnames = self._get_images_filenames()
                self._add_bright_pcount(exposure, photons, fnames)

                continue

            if l[0] == 'd':
                if len(l) != 2:
                    raise SyntaxError('Wrong format: "%s" should be "d '
                                      'exposure"' % line)

                exposure = np.float(l[1].replace(',', '.'))
                fnames = self._get_images_filenames()
                self._add_dark_pcount(exposure, fnames)

                continue

            # Data for datasheet operation point starts with o
            if l[0] == 'o':
                if len(l) < 2:
                    raise SyntaxError('Wrong format: %s should be "l varname '
                                      'values"' % line)
                self.info['operation_point_info'][l[1]] = ' '.join(l[2:])
                continue

            # Data for datasheet camera starts with c
            if l[0] == 'c':
                if len(l) < 2:
                    raise SyntaxError('Wrong format: %s should be "l varname '
                                      'values"' % line)
                self.info['camera_info'][l[1]] = ' '.join(l[2:])
                continue

            self.log.warning('Unknown command ' + line)

    def _split_line(self, line):
        '''
        For every line of descriptorfile
        check that it has at least two arguments
        split by white spaces and strip white spaces from elements
        '''
        l = [x.strip() for x in line.split()]
        if (not l) or (len(l) < 2):
            raise SyntaxError('Wrong format line: %s' % line)
        return l

    def _load_file(self, filename):
        '''
        Load a file, split by lines removing the comments (starts with #)
        '''
        self.log.info('Opening ' + filename)
        f = open(filename, 'r')

        # To add location when opening images
        path = os.path.realpath(filename)
        self.info['path'] = os.path.dirname(path)
        self.info['filename'] = os.path.basename(path)

        self._lines = [x.strip() for x in f.readlines() if x.strip() and
                       not x.strip().startswith('#')]



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Load image data
This class takes a dictionary (product of parser.ParseEmvaDescriptorFile).
Load the related images and reduce it's data to the minimum possible,
preserving all relevant image data in as integers

"""

import os
import numpy as np
from emva1288 import routines
import logging


class LoadImageData(object):
    '''Using a info dictionary, load the images and fill data with
    the relevant information

    Parameters
    ----------
    info: Dict result of ParseEmvaDescriptorFile
    loglevel: Logging level
    fload: function used to load the images
    fload_args: arguments passed to fload
    fload_kwargs: keyword arguments passed to fload
    path: path to load the images from, if None, the path in info is used
    '''

    def __init__(self,
                 info,
                 loglevel=logging.INFO,
                 fload=routines.load_image,
                 fload_args=[],
                 fload_kwargs={},
                 path=None):

        self.data = {'version': None,
                     'format': {},  # bits, witdth, height
                     'name': None,
                     'info': {},
                     'temporal': {'dark': {}, 'bright': {}},
                     'spatial': {'dark': {}, 'bright': {}},
                     }
        self._fload = fload
        self._fload_args = fload_args
        self._fload_kwargs = fload_kwargs

        logging.basicConfig()
        self.log = logging.getLogger('Loader')
        self.log.setLevel(loglevel)

        if path is not None:
            self._path = path
        else:
            self._path = info['path']
        self._load_data(info)

    def _load_data(self, info):
        '''Using the information in self.info fill self.data loading the images
        and filling the temporal and spatial dicts
        '''

        self.data['version'] = info['version']
        self.data['format'] = info['format']
        self.data['name'] = info['filename']
        self.data['camera_info'] = info['camera_info']
        self.data['operation_point_info'] = info['operation_point_info']

        for kind in ('temporal', 'spatial'):
            b_exp = set(info[kind]['bright'].keys())
            d_exp = set(info[kind]['dark'].keys())

            if b_exp - d_exp:
                raise SyntaxError('%s Bright and dark must have the '
                                  'same exposures' % kind)

            for exposure in b_exp:
                for photons in info[kind]['bright'][exposure]:
                    fnames = info[kind]['bright'][exposure][photons]
                    data_imgs = self._get_imgs_data(fnames, kind)
                    self.data[kind]['bright'].setdefault(exposure, {})
                    self.data[kind]['bright'][exposure][photons] = data_imgs

                fnames = info[kind]['dark'][exposure]
                data_imgs = self._get_imgs_data(fnames, kind)
                self.data[kind]['dark'][exposure] = data_imgs

    def _get_imgs_data(self, fnames, kind):
        '''Return the desired image data
        This depends on the kind of data
        '''
        arr_imgs = self._load_imgs(fnames)
        imgs = routines.get_int_imgs(arr_imgs)
        # For spatial we want the images
        if kind != 'temporal':
            return imgs
        # For temporal, we want numbers
        d = {}
        d['sum'] = np.sum(imgs['sum'])
        d['pvar'] = np.sum(imgs['pvar'])
        return d

    def _load_imgs(self, fnames):
        '''For a list of images, load and append them to the return array
        '''
        imgs = []
        for fname in fnames:
            filename = os.path.join(self._path, fname)
            self.log.debug('Loading ' + fname)
            if not os.path.isfile(filename):
                raise IOError('Not such file: ' + filename)
            imgs.append(self._fload(filename,
                                    *self._fload_args,
                                    **self._fload_kwargs))

        return imgs



import jinja2
import os
import shutil
from distutils.dir_util import copy_tree
from collections import namedtuple
from tempfile import TemporaryDirectory

from .. results import Results1288
from .. plotting import Plotting1288, EVMA1288plots


def marketing(**kwargs):
    m = namedtuple('marketing',
                   'logo, vendor, model, '
                   'serial, sensor_type, sensor_name, '
                   'resolution_x, resolution_y, '
                   'sensor_diagonal, lens_mount, '
                   'shutter, overlap, readout_rate, '
                   'dark_current_compensation, interface, '
                   'watermark, qe_plot')

    # For these attributes default is False
    # for the rest is '-'
    kwargs.setdefault('logo', False)
    kwargs.setdefault('watermark', False)
    kwargs.setdefault('qe_plot', False)
    for field in m._fields:
        v = kwargs.pop(field, '-')
        setattr(m, field, v)
    return m


def op():
    o = namedtuple('op', 'bit_depth, gain,'
                   'offset, exposure_time, wavelength, '
                   'temperature, housing_temperature, '
                   'fpn_correction, results, plots, extra')

    o.bit_depth = '-'
    o.gain = '-'
    o.offset = '-'
    o.exposure_time = '-'
    o.wavelength = '-'
    o.temperature = 'room'
    o.housing_temperature = '-'
    o.fpn_correction = '-'
    o.extra = False
    o.results = None
    o.plots = None
    o.id = None
    return o

_CURRDIR = os.path.abspath(os.path.dirname(__file__))


class Report1288(object):
    def __init__(self, mark):
        self._tmpdir = None

        self.renderer = jinja2.Environment(
            block_start_string='%{',
            block_end_string='%}',
            variable_start_string='%{{',
            variable_end_string='%}}',
            comment_start_string='%{#',
            comment_end_string='%#}',
            loader=jinja2.FileSystemLoader(os.path.join(_CURRDIR, 'templates'))
        )
        self.ops = []
        self.marketing = mark
        self._temp_dirs()

    def _temp_dirs(self):
        self._tmpdir = TemporaryDirectory()
        tmpfiles = os.path.join(self._tmpdir.name, 'files')
        os.makedirs(tmpfiles)
        currfiles = os.path.join(_CURRDIR, 'files')
        copy_tree(currfiles, tmpfiles)
        markfiles = os.path.join(self._tmpdir.name, 'marketing')
        os.makedirs(markfiles)

        if self.marketing.logo:
            fname = os.path.basename(self.marketing.logo)
            if not fname:
                print('invalid logo file', fname)
            else:
                shutil.copy(os.path.abspath(self.marketing.logo),
                            os.path.join(markfiles, fname))
                self.marketing.logo = os.path.join('marketing', fname)

        if self.marketing.qe_plot:
            fname = os.path.basename(self.marketing.qe_plot)
            if not fname:
                print('invalid qe_plot file', fname)
            else:
                shutil.copy(os.path.abspath(self.marketing.qe_plot),
                            os.path.join(markfiles, fname))
                self.marketing.qe_plot = os.path.join('marketing', fname)

    def _write_file(self, name, content):
        fname = os.path.join(self._tmpdir.name, name)
        with open(fname, 'w') as f:
            f.write(content)
        return fname

    def _stylesheet(self):
        stylesheet = self.renderer.get_template('emvadatasheet.sty')
        return stylesheet.render(marketing=self.marketing)

    def _report(self):
        report = self.renderer.get_template('report.tex')
        return report.render(marketing=self.marketing,
                             operation_points=self.ops)

    def latex(self, dir_):
        '''Generate report latex files in a given directory'''

        self._write_file('emvadatasheet.sty', self._stylesheet())
        self._write_file('report.tex', self._report())

        outdir = os.path.abspath(dir_)
        try:
            os.makedirs(outdir)
        except FileExistsError:
            pass
        copy_tree(self._tmpdir.name, outdir)
        print('Report files found in:', outdir)

    def _results(self, data):
        return Results1288(data)

    def _plots(self, data, id_):
        res = self._results(data)
        plots = Plotting1288(res)
        savedir = os.path.join(self._tmpdir.name, id_)
        os.mkdir(savedir)
        plots.plot(savedir=savedir, show=False)
        names = {}
        for cls in EVMA1288plots:
            names[cls.__name__] = os.path.join(id_, cls.__name__ + '.pdf')
        return names

    def add(self, op_, data=None):
        if not op_.id:
            op_.id = 'OP%d' % (len(self.ops) + 1)
        if not op_.results and data:
            op_.results = self._results(data).results
        if not op_.plots and data:
            op_.plots = self._plots(data, op_.id)
        self.ops.append(op_)



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Utils functions

"""
from __future__ import print_function
import numpy as np
import os
from scipy.optimize import leastsq
from lxml import etree
from PIL import Image


SIGNIFICANT_DIGITS = 7


def load_image(fname):
    img = Image.open(fname)
    img = np.asarray(img.split()[0])
#     img = cv2.imread(fname, cv2.CV_LOAD_IMAGE_UNCHANGED)
#     img = cv2.split(img)[0]
    return img


def get_int_imgs(imgs):
    '''
    Returns the sum and pseudo-variance from list of images
    sum is just the image resulting on the addition of all the images
    pvar is the pseudo-variance, this is
    pvar = SUM((Li - SUM(i))^2)
    to get variance from pseudo-variance
    var = (1/(L^2) * 1/(L - 1)) * pvar
    '''
    L = len(imgs)

    sum_ = 0
    sq_ = 0
    for img in imgs:
        # we force the images as int64 to make sure we do not clip
        i = img.astype(np.int64)
        sum_ += i
        sq_ += np.square(i)

    # the pseudo variance can be computed from the sum image and the sum of
    # the square images
    var_ = L * (L * sq_ - np.square(sum_))

    return {'L': L, 'sum': sum_, 'pvar': var_}


def LinearB0(Xi, Yi):
    X = np.asfarray(Xi)
    Y = np.asfarray(Yi)

    # we want a function y = m * x
    def fp(v, x):
        return x * v[0]

    # the error of the function e = x - y
    def e(v, x, y):
        return (fp(v, x) - y)

    # the initial value of m, we choose 1, because we thought YODA would
    # have chosen 1
    v0 = [1.0]

    vr, _success = leastsq(e, v0, args=(X, Y))

    # compute the R**2 (sqrt of the mean of the squares of the errors)
    err = np.sqrt(sum(np.square(e([vr], X, Y))) / (len(X) * len(X)))

    # Some versions of leastsq returns an array, other a scalar, so here we
    # make sure
    # it is an array
    val = np.array([vr]).flatten()

    return val, err


def LinearB(Xi, Yi):
    X = np.asfarray(Xi)
    Y = np.asfarray(Yi)

    # we want a function y = m * x
    def fp(v, x):
        return x * v[0] + v[1]

    # the error of the function e = x - y
    def e(v, x, y):
        return (fp(v, x) - y)

    # the initial value of m, we choose 1, because we thought YODA would
    # have chosen 1
    v0 = np.array([1.0, 1.0])

    vr, _success = leastsq(e, v0, args=(X, Y))

    # compute the R**2 (sqrt of the mean of the squares of the errors)
    err = np.sqrt(sum(np.square(e(vr, X, Y))) / (len(X) * len(X)))

#    print vr, success, err
    return vr, err


def GetImgShape(img):
    rows = 1

    if img.ndim == 1:
        cols, = img.shape
    else:
        rows, cols = img.shape
    return rows, cols


def FFT1288(m, rotate=False):
    mm = np.asfarray(np.copy(m))

    if rotate is True:
        mm = mm.transpose()

    _rows, cols = GetImgShape(mm)

    # This is just in case we are talking about really small or really
    # big arrays
    if (cols < 10) or (cols > 50000):
        return []

    # Substract the mean of the image
    mm = mm - np.mean(mm)

    # perform the fft in the x direction
    fft = np.fft.fft(mm, axis=1)

    fft = fft / np.sqrt(cols)

    fabs = np.real(fft * np.conjugate(fft))

    # extract the mean of each column of the fft
    r = np.mean(fabs, axis=0)

    return r


def GetFrecs(fft):
    n = len(fft)
    x = np.arange(n)
    x = x * 1.0 / (2 * n)
    return x


def Histogram1288(img, Qmax):
    y = np.ravel(img)

    ymin = np.min(y)
    ymax = np.max(y)

    # Because we are working with integers, minimum binwidth is 1
    W = 1
    q = ymax - ymin
    Q = q + 1

    # When too many bins, create a new integer binwidth
    if Q > Qmax:
        # We want the number of bins as close as possible to Qmax (256)
        W = int(np.ceil(1. * q / (Qmax - 1)))
        Q = int(np.floor(1. * q / W)) + 1

    # The bins
    # we need one more value for the numpy histogram computation
    # numpy used bin limits
    # in our interpretation we use the lower limit of the bin
    B = [ymin + (i * W) for i in range(Q + 1)]

    # Normal distribution with the original sigma, and mean
    mu = np.mean(y)
    sigma = np.std(y)
    normal = ((1. * (ymax - ymin) / Q) *
              np.size(y) / (np.sqrt(2 * np.pi) * sigma) *
              np.exp(-0.5 * (1. / sigma * (B[:-1] - mu)) ** 2))


#############################################
#
#   # Reference algorithm, it's pretty slow
#   # the numpy version gives the same results
#
#     #The histogram container
#     H = np.zeros((Q,), dtype=np.int64)
#
#     #Histogram computation
#     for yi in y:
#         q = (yi - ymin) / W
#         H[q] += 1
#############################################

    H, _b = np.histogram(y, B, range=(ymin, ymax))

    return {'bins': np.asfarray(B[:-1]), 'values': H, 'model': normal}


def cls_1288_info(cls):
    '''
    Returns a dict "d" of the form
    d = {SectionName : {MethodName: {TagName: TagValue, ...}}}
    TagName and TagValue are pairs in the docstring with the format
    **Tagname: TagValue

    SectionName correspond a tag **Section: SectionName

    If there is no such tag, the SectionName used for the method is 'other'
    '''

    d = {}
    for i, v in cls.__dict__.items():
        # Extract the doc from the Processing methods
        doc = v.__doc__
        if not doc:
            continue

        # All the lines in the docstring
        lines = [s.strip() for s in doc.splitlines()]

        # to store the relevant tag lines
        tag_lines = []

        for line in lines:
            # Get only those that are relevant (start with **)
            if line.startswith('**'):
                tag_lines.append(line[2:])
        # if there are not relevant lines skip and go to next method
        if not tag_lines:
            continue

        # To store the info from the doc
        method_info = {}
        for line in tag_lines:
            tags = [x.strip() for x in line.split(':')]
            # Each valid tag has to be xx:yy
            if len(tags) != 2:
                continue
            # Fill the dict
            method_info[tags[0]] = tags[1]

        # extract the section, or put it as 'other'
        section = method_info.pop('Section', 'other')

        # Add or get the current section branch of the dict
        s = d.setdefault(section, {})
        # Add the method info to the section in the final dict
        s[i] = method_info
    return d


def obj_to_dict(obj):
    '''
    Get the info dict from the object class
    Add the values or Data to this dict

    for each method if the return value is a dict, it is inserted as
    d[SectionName][MethodName][Data] = ReturnValue
    if not
    d[SectionName][MethodName][Value] = ReturnValue
    '''
    d = cls_1288_info(obj.__class__)
    for section in d.keys():
        for methodname in d[section].keys():

            # Get the value for the given method
            val = getattr(obj, methodname)
            if callable(val):
                val = val()

            if isinstance(val, dict):
                d[section][methodname]['Data'] = val
            else:
                d[section][methodname]['Value'] = val
    return d


def dict_to_xml(d, root='results', filename=None):
    '''
    Takes a dict and return a well formed xml string
    '''

    def key_to_xml(d, r):
        '''
        Recursive call to add the key/value from dict to the r element tree
        If the value is an array joint the values casted as strings
        with whitespaces separator
        if the value is something else, it is casted as string
        '''
        for k in d.keys():
            e = etree.SubElement(r, k)
            # when the value for the key is a dict, call the function again
            if isinstance(d[k], dict):
                r.append(key_to_xml(d[k], e))

            # if the value is an array
            # add the values of the array as a string separated by whitespaces
            # Note to self: add other array types as needed
            elif isinstance(d[k], np.ndarray):
                a = [str(x) for x in d[k]]
                e.text = ' '.join(a)

            # if something else, just add the corresponding string value
            else:
                e.text = str(d[k])
        return r

    tree = etree.Element(root)
    xml = key_to_xml(d, tree)

    t = etree.tostring(xml, pretty_print=True)
    if filename is None:
        return t
    filename.write(t)


def xml_to_dict(xml):
    '''
    If xml is a file, opens and parse, if string, parse it from string
    Convert the xml to a dict using element_to_dict
    Process the resulting dict:
    Cast Data to numpy float arrays (split the string by whitespaces)
    Cast Value to float
    '''
    try:
        if os.path.isfile(xml):
            tree = etree.parse(xml)
        else:
            tree = etree.ElementTree.fromstring(xml)
    except:
        print('Problems loading XML')
        return None

    def element_to_dict(r):
        '''
        Recursive call to add dictionnary elements from the r xml element
        '''
        dout = {}
        for child in r:
            if list(child):
                # if the element has children call the function again with
                # children as r
                dout[child.tag] = element_to_dict(child)
            else:
                dout[child.tag] = child.text
        return dout

    root = tree.getroot()
    d = element_to_dict(root)
    # loop to reconstruct arrays from strings in Data elements
    for section, method in d.items():
        for methodname, value in method.items():
            if 'Data' in value:
                for data in value['Data']:
                    v = value['Data'][data]
                    v = v.strip()
                    v = v.split()
                    d[section][methodname]['Data'][data] = np.asfarray(v)
            else:
                # sometimes the decimal point is written with , instead of .
                v = value['Value'].replace(',', '.')
                # special cases, None, etc...
                if v == 'None':
                    v = None
                else:
                    v = float(v)
                d[section][methodname]['Value'] = v
    return d


def round_significant(v, sig=SIGNIFICANT_DIGITS):
    '''
    Round up to the given significant digits, used for comparison
    '''
    if v == 0.0:
        return 0.0
    return round(v, sig - np.int(np.floor(np.log10(np.abs(v)))) - 1)


round_array = np.vectorize(round_significant)


def compare_xml(x1, x2):
    # load the xml into dicts
    f1 = xml_to_dict(x1)
    f2 = xml_to_dict(x2)

    # if something is wrong abort
    if f1 is None or f2 is None:
        return

    c1 = f1.keys()
    c2 = f2.keys()

    # loop throught the combined categories
    for category in set(c1) | set(c2):
        print('')
        print('*' * 70)
        print(category)
        print('*' * 70)
        # check if missing category in one of the dicts
        if category not in c1 or category not in c2:
            t1 = category in c1
            t2 = category in c2
            print('{0:<35}'.format('PRESENT'), end=" ")
            print('{0:<20}{1:<20}FAIL'.format(str(t1), str(t2)))
            continue

        m1 = f1[category].keys()
        m2 = f2[category].keys()
        # loop throught the combined methodnames
        for methodname in set(m1) | set(m2):
            print('{0:<35}'.format(methodname), end=" ")

            # check if methodname in dict
            if methodname not in m1:
                v1 = None
                a2 = None
            # get the value and the data
            else:
                v1 = f1[category][methodname].get('Value', None)
                a1 = f1[category][methodname].get('Data', None)

            if methodname not in m2:
                v2 = None
                a2 = None
            else:
                v2 = f2[category][methodname].get('Value', None)
                a2 = f2[category][methodname].get('Data', None)

            # first we check for values and then for data
            # if both present only values will be taken in count for comparison
            # If both are values
            if v1 is not None and v2 is not None:
                try:
                    r = (round_significant(v1) - round_significant(v2)) == 0.0
                except:
                    r = False
                t1 = v1
                t2 = v2

            # if both are arrays
            elif a1 is not None and a2 is not None:
                k1 = a1.keys()
                k2 = a2.keys()
                t1 = 'Array'
                t2 = 'Array'

                # if different keys, is invalid
                if (set(k1) ^ set(k2)):
                    r = False
                else:
                    # loop throught the keys
                    for k in k1:
                        try:
                            r = np.max(np.abs(round_array(a1[k]) -
                                              round_array(a2[k]))) == 0.0
                        except:
                            r = False

                        if not r:
                            break

            print('{0:<20}{1:<20}'.format(t1, t2), end=" ")
            if r:
                print('OK')
            else:
                print('FAIL')



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Compute EMVA1288 values from data
This class takes the data from data.Data1288 and compute the actual EMVA1288
values.

"""

from __future__ import print_function
import logging
import numpy as np
from emva1288 import routines
from scipy.ndimage import convolve


class Results1288(object):
    def __init__(self,
                 data,
                 index_u_ysat=None,
                 loglevel=logging.DEBUG):

        logging.basicConfig()
        self.log = logging.getLogger('Results')
        self.log.setLevel(loglevel)

        self.temporal = data['temporal']
        self.spatial = data['spatial']
        self.name = data['name']
        self._s2q = 1.0 / 12.0
        self._index_start = 0
        self._index_sensitivity_min = 0

        self._histogram_Qmax = 256  # Maximum number of bins in histograms
        # Convolution kernel for high pass filter in defect pixel
        self._histogram_high_pass_box = (-1) * np.ones((5, 5), dtype=np.int64)
        self._histogram_high_pass_box[2, 2] = 24

        # Sometimes we need to force the saturation point
        # in those cases pass the index in the initialization of Results1288
        self._index_u_ysat = index_u_ysat

    @property
    def s2q(self):
        return self._s2q

    @property
    def index_start(self):
        """
        **Section: info
        **Short: The array indexes start at
        **Symbol:
        **Comment:
        **Unit:
        """
        return self._index_start

    @property
    def index_u_ysat(self):
        """
        **Section: sensitivity
        **Short: Saturation Point
        **Symbol:
        **Comment: Index of saturation
        **Unit:
        """

        if self._index_u_ysat:
            return self._index_u_ysat

        max_ = 0
        max_i = 0

        # we have to loop backwards because sometimes we have some
        # noise pics that really bother the computation
        s2_y = self.temporal['s2_y']
        for i in range(len(s2_y) - 1, -1, -1):
            # Check that is not a local max
            if (s2_y[i] >= max_) or \
               (s2_y[abs(i - 1)] >= max_):
                max_ = s2_y[i]
                max_i = i
            elif (s2_y[i] < max_) and \
                 (s2_y[abs(i - 1)] < max_):
                break

        self._index_u_ysat = max_i
        return self._index_u_ysat

    @property
    def index_sensitivity_max(self):
        """Index for linear fits in sensitivity part of the standard
        (70% of saturation)
        **Section: sensitivity
        **Symbol:
        **Unit:
        **Short:Sensitivity fit maximum index
        """

        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        m = 0.7 * (Y[self.index_u_ysat])
        return max(np.argwhere(Y <= m))[0]

    @property
    def index_sensitivity_min(self):
        """
        **Section: sensitivity
        **Short:Sensitivity fit minimum index
        **Symbol:
        **Comment:Index for linear fits in sensitivity part of the standard (70% of saturation)
        **Unit:
        """
        return self._index_sensitivity_min

    @property
    def R(self):
        """
        **Section: sensitivity
        **Short:Responsivity
        **Symbol:R
        **Comment:Slope of the (u_y - u_y_dark) Vs u_p. Fit with offset = 0
        **Unit:DN/p
        """

        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        X = self.temporal['u_p']

        val, _error = routines.LinearB0(X[self.index_sensitivity_min:
                                          self.index_sensitivity_max + 1],
                                        Y[self.index_sensitivity_min:
                                          self.index_sensitivity_max + 1])

        return val[0]

    @property
    def K(self):
        """
        **Section: sensitivity
        **Short:System gain
        **Symbol:K
        **Comment:Slope of (s2_y - s2_y_dark) Vs (u_y - u_y_dark). Fit with offset = 0
        **Unit: $DN/e^-$
        **LatexName:K
        """

        X = self.temporal['u_y'] - self.temporal['u_ydark']
        Y = self.temporal['s2_y'] - self.temporal['s2_ydark']

        val, _error = routines.LinearB0(X[self.index_sensitivity_min:
                                          self.index_sensitivity_max + 1],
                                        Y[self.index_sensitivity_min:
                                          self.index_sensitivity_max + 1])

        return val[0]

    def inverse_K(self):
        """
        **Section: sensitivity
        **Short:Inverse system gain
        **Symbol: 1/K
        **Comment:
        **Unit: $e^-/DN$
        **LatexName: InvK
        """

        return 1. / self.K

    @property
    def QE(self):
        """
        **Section: sensitivity
        **Short: Quantum efficiency
        **Symbol: $\eta$
        **Comment:
        **Unit: \%
        **LatexName: QE
        """

        return 100.0 * self.R / self.K

    @property
    def sigma_y_dark(self):
        """
        **Section: sensitivity
        **Short: Temporal Dark Noise
        **Symbol: $\sigma_{y.dark}$
        **Comment:
        **Unit: DN
        **LatexName: SigmaYDark
        """

        if len(np.unique(self.temporal['texp'])) <= 2:
            s2_ydark = self.temporal['s2_ydark'][0]
        else:
            fit, _error = routines.LinearB(self.temporal['texp'],
                                           self.temporal['s2_ydark'])
            s2_ydark = fit[1]

        # Lower limit for the temporal dark noise
        # The temporal dark noise in this range is dominated by the
        # quantization noise
        if s2_ydark < 0.24:
            s2_ydark = 0.24

        return np.sqrt(s2_ydark)

    @property
    def sigma_d(self):
        """
        **Section: sensitivity
        **Short: Temporal Dark Noise
        **Symbol:$\sigma_d$
        **Comment:
        **Unit: $e^-$
        **LatexName: SigmaDark
        """

        return np.sqrt((self.sigma_y_dark ** 2) - self._s2q) / self.K

    @property
    def u_p_min(self):
        """
        **Section: sensitivity
        **Short: Sensitivity threshold
        **Symbol:$\mu_{p.min}$
        **Comment:
        **Unit: $p$
        **LatexName: UPMin
        """

        return (100.0 / self.QE) * ((self.sigma_y_dark / self.K) + 0.5)

    def u_e_min(self):
        """
        **Section: sensitivity
        **Short: Sensitivity threshold
        **Symbol:$\mu_{e.min}$
        **Comment:
        **Unit: $e^-$
        **LatexName: UEMin
        """
        return self.QE * self.u_p_min / 100.0

    @property
    def u_p_sat(self):
        """
        **Section: sensitivity
        **Short:Saturation Capacity
        **Symbol:$\mu_{p.sat}$
        **Comment:
        **Unit: $p$
        **LatexName: UPSat
        """

        return self.temporal['u_p'][self.index_u_ysat]

    @property
    def u_e_sat(self):
        """
        **Section: sensitivity
        **Short:Saturation Capacity
        **Symbol:$\mu_{e.sat}$
        **Comment:Number of electrons at saturation
        **Unit: $e^-$
        **LatexName: UESat
        """

        return self.QE * self.u_p_sat / 100.0

    @property
    def SNR_max(self):
        """
        **Section: sensitivity
        **Short: Maximum Signal to Noise Ratio
        **Symbol: $SNR_{max}$
        **Comment:
        **Unit:
        **LatexName: SNRMax
        """

        return np.sqrt(self.u_e_sat)

    def SNR_max_dB(self):
        """
        **Section: sensitivity
        **Short:Maximum Signal to Noise Ratio in Db
        **Symbol: $SNR_{max.dB}$
        **Comment:
        **Unit: dB
        **LatexName: SNRMaxDB
        """

        return 20. * np.log(self.SNR_max)

    def SNR_max_bit(self):
        """
        **Section: sensitivity
        **Short: Maximum Signal to Noise Ratio in Bits
        **Symbol: $SNR_{max.bit}$
        **Comment:
        **Unit: bit
        **LatexName: SNRMaxBit
        """

        return np.log2(self.SNR_max)

    def inverse_SNR_max(self):
        """
        **Section: sensitivity
        **Short: Inverse Maximum Signal to Noise Ratio
        **Symbol: $SNR_{max}^{-1}$
        **Comment:
        **Unit: \%
        **LatexName: InvSNRMax
        """

        return 100. / self.SNR_max

    @property
    def DR(self):
        """
        **Section: sensitivity
        **Short: Dynamic Range
        **Symbol: DR
        **Comment:
        **Unit:
        **LatexName: DR
        """

        return self.u_p_sat / self.u_p_min

    def DR_dB(self):
        """
        **Section: sensitivity
        **Short: Dynamic Range
        **Symbol: $DR_{dB}$
        **Comment:
        **Unit: dB
        **LatexName: DRDB
        """

        return 20. * np.log10(self.DR)

    def DR_bit(self):
        """
        **Section: sensitivity
        **Short: Dynamic Range
        **Symbol: $DR_{bit}$
        **Comment:
        **Unit: bit
        **LatexName: DRBit
        """

        return np.log2(self.DR)

    @property
    def index_linearity_min(self):
        """
        **Section: linearity
        **Short:Linearity fit minimun index
        **Symbol:
        **Comment:Minimum index for linear fit in (5% of saturation)
        **Unit:
        """
        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        vmin = 0.05 * (Y[self.index_u_ysat])

        return min(np.argwhere(Y >= vmin))[0]

    @property
    def index_linearity_max(self):
        """
        **Section: linearity
        **Short: Linearity fit maximum index
        **Symbol:
        **Comment:
        **Unit:
        """

        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        vmax = 0.95 * (Y[self.index_u_ysat])
        return max(np.argwhere(Y <= vmax))[0]

    def linearity(self):
        Y = self.temporal['u_y'] - self.temporal['u_ydark']
        X = self.temporal['u_p']

        # The maximum index has to be included, this is the reason for the +1
        imax = self.index_linearity_max + 1
        imin = self.index_linearity_min

        X_ = X[imin: imax]
        Y_ = Y[imin: imax]
        xy = np.sum(X_ / Y_)
        xy2 = np.sum(X_ / (Y_ ** 2))
        x2y2 = np.sum((X_ / Y_) ** 2)
        _y = np.sum(1. / Y_)
        _y2 = np.sum(1. / (Y_ ** 2))

        b = ((xy * xy2) - (x2y2 * _y)) / ((xy2 ** 2) - (x2y2 * _y2))
        a = (xy - (b * xy2)) / x2y2

        dev = 100. * (Y - (a * X + b)) / (a * X + b)

        lin = {}
        lin['fit_slope'] = a
        lin['fit_offset'] = b
        lin['relative_deviation'] = dev
        lin['linearity_error_min'] = np.min(dev[imin: imax])
        lin['linearity_error_max'] = np.max(dev[imin: imax])

        return lin

    @property
    def LE_min(self):
        """
        **Section: linearity
        **Short: Min Linearity error
        **Symbol: $LE_{min}$
        **Comment:
        **Unit: \%
        **LatexName: LEMin
        """

        return self.linearity()['linearity_error_min']

    @property
    def LE_max(self):
        """
        **Section: linearity
        **Short: Max Linearity error
        **Symbol: $LE_{max}$
        **Comment:
        **Unit: \%
        **LatexName: LEMax
        """

        return self.linearity()['linearity_error_max']

    @property
    def u_I_var(self):
        """
        **Section: dark_current
        **Short:Dark Current from variance
        **Symbol:$\mu_{I.var}$
        **Comment:
        **Unit: $e^-/s$
        **LatexName: UIVar
        """

        fit, _error = routines.LinearB(self.temporal['texp'],
                                       self.temporal['s2_ydark'])

        if fit[0] < 0:
            return np.nan
        return np.sqrt(fit[0] * (10 ** 9)) / self.K

    @property
    def u_I_mean(self):
        """
        **Section: dark_current
        **Short:Dark Current from mean
        **Symbol:$\mu_{I.mean}$
        **Comment:
        **Unit: e/s
        """

        if len(np.unique(self.temporal['texp'])) <= 2:
            return np.nan

        fit, _error = routines.LinearB(self.temporal['texp'],
                                       self.temporal['u_ydark'])
        return fit[0] * (10 ** 9) / self.K

    @property
    def sigma_2_y_stack(self):
        """
        **Section: spatial
        **Short: Temporal variance stack
        **Symbol: $\sigma^2_{y.stack}$
        **Comment: Mean value of the bright variance image
        **Unit: DN2
        """

        return np.mean(self.spatial['var'][0])

    @property
    def sigma_2_y_stack_dark(self):
        """
        **Section: spatial
        **Short: Temporal variance stack dark
        **Symbol:$\sigma^2_{y.stack.dark}$
        **Comment:Mean value of the dark variance image.
        **Unit:DN2
        """

        return np.mean(self.spatial['var_dark'][0])

    @property
    def s_2_y_measured(self):
        """
        **Section: spatial
        **Short: Spatial variance measure
        **Symbol:$s^2_{y.measured}$
        **Comment:Variance value of the bright variance image
        **Unit:DN2
        """

        return np.var(self.spatial['avg'][0], ddof=1)

    @property
    def s_2_y_measured_dark(self):
        """
        **Section: spatial
        **Short: Spatial variance measured dark
        **Symbol:$s^2_{y.measured.dark}$
        **Comment:Variance value of the dark variance image
        **Unit:DN2
        """

        return np.var(self.spatial['avg_dark'][0], ddof=1)

    @property
    def s_2_y(self):
        """
        **Section: spatial
        **Short: Spatial variance from image
        **Symbol:$s^2_{y}$
        **Comment:s_2_y_measured - sigma_2_y_stack / number_images
        **Unit:DN2
        """

        return self.s_2_y_measured - (self.sigma_2_y_stack /
                                      self.spatial['L'][0])

    @property
    def s_2_y_dark(self):
        """
        **Section: spatial
        **Short: Spatial variance from image
        **Symbol:$s^2_{y}$
        **Comment:s_2_y_measured - sigma_2_y_stack / number_images
        **Unit:DN2
        """

        return self.s_2_y_measured_dark - (self.sigma_2_y_stack_dark /
                                           self.spatial['L_dark'][0])

    @property
    def s_2_y_measured_spectrogram(self):
        """
        **Section: spatial
        **Short: Spatial variance measured from spectrogram
        **Symbol:$s^2_{y.measured.spectrogram}$
        **Comment:
        **Unit:DN2
        """

        return np.mean(routines.FFT1288(self.spatial['avg'][0])[8: -7])

    @property
    def s_2_y_measured_spectrogram_dark(self):
        """
        **Section: spatial
        **Short: Spatial variance measured dark from spectrogram
        **Symbol:$s^2_{y.measured.spectrogram.dark}$
        **Comment:
        **Unit:DN2
        """

        return np.mean(routines.FFT1288(self.spatial['avg_dark'][0]))

    @property
    def s_2_y_spectrogram(self):
        """
        **Section: spatial
        **Short: Spatial variance from spectrogram
        **Symbol:$s^2_{y.spectrogram}$
        **Comment:
        **Unit:DN2
        """
        return (self.s_2_y_measured_spectrogram -
                (self.sigma_2_y_stack / self.spatial['L'][0]))

    @property
    def s_2_y_spectrogram_dark(self):
        """
        **Section: spatial
        **Short: Spatial variance from spectrogram dark
        **Symbol:$s^2_{y.spectrogram.dark}$
        **Comment:
        **Unit:DN2
        """

        return (self.s_2_y_measured_spectrogram_dark -
                (self.sigma_2_y_stack_dark / self.spatial['L_dark'][0]))

    @property
    def DSNU1288(self):
        """
        **Section: spatial
        **Short: DSNU
        **Symbol: $DSNU_{1288}$
        **Comment:
        **Unit: $e^-$
        **LatexName: DSNU
        """

        if self.s_2_y_dark < 0.:
            return np.nan
        return np.sqrt(self.s_2_y_dark) / self.K

    def DSNU1288_DN(self):
        """
        **Section: spatial
        **Short: DSNU in DN
        **Symbol: $DSNU_{1288.DN}$
        **Comment:
        **Unit: DN
        **LatexName: DSNUDN
        """

        if self.s_2_y_dark < 0:
            return np.nan
        return np.sqrt(self.s_2_y_dark)

    @property
    def PRNU1288(self):
        """
        **Section: spatial
        **Short: PRNU
        **Symbol: $PRNU_{1288}$
        **Comment:
        **Unit: \%
        **LatexName: PRNU
        """

        return (np.sqrt(self.s_2_y - self.s_2_y_dark) * 100 /
                (np.mean(self.spatial['avg'][0]) -
                 np.mean(self.spatial['avg_dark'][0])))

    @property
    def F_50(self):
        """
        **Section: spatial
        **Short: Non Whiteness factor 50%
        **Symbol: $F_{50\%}$
        **Comment:
        **Unit:
        """

        return (self.s_2_y_measured_spectrogram /
                np.median(routines.FFT1288(self.spatial['avg'][0])[8: -7]))

    @property
    def F_dark(self):
        """
        **Section: spatial
        **Short: Non Whiteness factor Dark
        **Symbol: $F_{dark}$
        **Comment:
        **Unit:
        """

        return (self.s_2_y_measured_spectrogram_dark /
                np.median(routines.FFT1288(self.spatial['avg_dark'][0])))

    @property
    def histogram_PRNU(self):
        """
        **Section: defect_pixel
        **Short: PRNU histogram
        **Symbol:
        **Comment:
        **Unit:
        """

        # For prnu, perform the convolution
        y = self.spatial['sum'][0] - self.spatial['sum_dark'][0]
        y = convolve(y, self._histogram_high_pass_box)[2:-2, 2:-2]

        h = routines.Histogram1288(y, self._histogram_Qmax)
        # Rescale the bins
        h['bins'] /= (self.spatial['L'][0] * 25.)

        return h

    @property
    def histogram_PRNU_accumulated(self):
        """
        **Section: defect_pixel
        **Short: accumulated PRNU histogram
        **Symbol:
        **Comment:
        **Unit:
        """

        # For prnu, perform the convolution
        y = self.spatial['sum'][0] - self.spatial['sum_dark'][0]
        y = convolve(y, self._histogram_high_pass_box)[2:-2, 2:-2]

        # For the accumulated histogram substract the mean
        y = np.abs(y - int(np.mean(y)))

        h = routines.Histogram1288(y, self._histogram_Qmax)
        # Rescale the bins
        h['bins'] /= (self.spatial['L'][0] * 25.)

        # Perform the cumulative summation
        h['values'] = np.cumsum(h['values'][::-1])[::-1]

        return h

    @property
    def histogram_DSNU(self):
        """
        **Section: defect_pixel
        **Short: DSNU histogram
        **Symbol:
        **Comment:
        **Unit:
        """

        # For dsnu, the image is just the dark image, upscaled to have
        # only integers
        y = self.spatial['sum_dark'][0]

        h = routines.Histogram1288(y, self._histogram_Qmax)
        # Rescale the bins, this is due to upscaling the average image to have
        # only integers
        h['bins'] /= (self.spatial['L_dark'][0] * 25.)

        return h

    @property
    def histogram_DSNU_accumulated(self):
        """
        **Section: defect_pixel
        **Short: accumulated DSNU histogram
        **Symbol:
        **Comment:
        **Unit:
        """

        # Dark image upscaled to have only integers
        y = self.spatial['sum_dark'][0]
        # For the accumulated dsnu histogram, substract the mean from the image
        y = np.abs(y - int(np.mean(y)))

        h = routines.Histogram1288(y, self._histogram_Qmax)
        # Rescale the bins
        h['bins'] /= (self.spatial['L_dark'][0] * 25.)

        # Perform the cumulative summation (the ::-1 means backwars)
        # because the cumsum function is performed contrary to what we need
        h['values'] = np.cumsum(h['values'][::-1])[::-1]

        return h

    def xml(self, filename=None):
        d = routines.obj_to_dict(self)
        if not filename:
            return routines.dict_to_xml(d)
        routines.dict_to_xml(d, filename=filename)

    @property
    def results(self):
        d = routines.obj_to_dict(self)
        results = {}
        for section in d.keys():
            results[section] = {}
            for method in d[section].keys():
                s = d[section][method]
                if s.get('Value', False) is not False:
                    results[section][method] = {'short': s['Short'],
                                                'symbol': s['Symbol'],
                                                'value': s['Value'],
                                                'unit': s['Unit'],
                                                'latexname': s.get('LatexName')}

        return results

    def print_results(self):
        d = self.results
        for section in d.keys():
            print('*' * 50)
            print(section)
            print('*' * 50)

            for method in d[section].keys():
                s = d[section][method]
                print('{:<50}{:<30}{:>10}'.format(s['short'],
                                                  s['symbol'],
                                                  s['value']))
        print('*' * 50)
        print(' ')



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""EMVA 1288 Package initialisation file"""

from .data import Data1288
from .loader import LoadImageData
from .parser import ParseEmvaDescriptorFile
from .results import Results1288
from .plotting import Plotting1288
from .report import report


class Emva1288(object):
    def __init__(self, fname):
        info = ParseEmvaDescriptorFile(fname)
        imgs = LoadImageData(info.info)
        dat = Data1288(imgs.data)
        self._results = Results1288(dat)

    def results(self):
        return self._results.print_results()

    def plot(self):
        plot = Plotting1288(self._results)
        plot.plot()



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Plot the results
This class takes a results.Results1288 object and produces all the plots
needed to create a reference datasheet of the EMVA1288 test

"""

from __future__ import print_function
# from matplotlib.figure import Figure
# from matplotlib.backend import new_figure_manager_given_figure
# import matplotlib.pyplot as plt
import numpy as np
import os

from emva1288 import routines


class Emva1288Plot(object):
    """Base class for emva plots

    The only mandatory attribute is name, the rest are for use in
    setup_figure
    """
    name = ""
    title = None
    xlabel = None
    ylabel = None
    xscale = None
    yscale = None

    def __init__(self, figure):
        self.figure = figure
        self.tests = []
        self.setup_figure()

    def setup_figure(self):
        """Simple wrapper for one plot per figure

        Takes the name, xlabel, ylabel, xscale and yscale
        for one plot case

        If more than one plot, just overwrite as you wish
        """
        ax = self.figure.add_subplot(111)
        if self.title:
            ax.set_title(self.title)
        else:
            ax.set_title(self.name)
        if self.xlabel:
            ax.set_xlabel(self.xlabel)
        if self.ylabel:
            ax.set_ylabel(self.ylabel)
        if self.xscale:
            ax.set_xscale(self.xscale)
        if self.yscale:
            ax.set_yscale(self.yscale)

        self.ax = ax

    def plot(self, test):
        pass

    def add_test(self, test):
        name = test.name
        if not name:
            name = len(self.tests)
        if not getattr(test, 'id', False):
            test.id = id(test)
        self.tests.append(name)
        self.plot(test)

    def set_legend(self, ax):
        '''
        This function is just to prevent a very long legend when there are
        several datasets
        If more than one test, only label the first plot as Dataset #
        if not then use the labels setted in the different plots
        '''

        handles, labels = ax.get_legend_handles_labels()
        if len(self.tests) > 1:
            d = int(len(labels) / len(self.tests))
            for j in range(len(labels)):
                t = int(j / d)
                labels[j] = 'Dataset: ' + str(self.tests[t])
            ax.legend(handles[::d], labels[::d], loc='upper left')
        else:
            ax.legend(loc='upper left')

        legend = ax.get_legend()
        if legend is not None:
            if getattr(legend, 'draggable', False):
                legend.draggable(True)


class PlotSensitivity(Emva1288Plot):
    name = 'Sensitivity'
    xlabel = '$\mu_p$ [mean number of photons/pixel]'
    ylabel = '$\mu_y - \mu_{y.dark}$ [DN]'

    def plot(self, test):
        ax = self.ax
        ax.plot(test.temporal['u_p'],
                test.temporal['u_y'] - test.temporal['u_ydark'],
                label='Data',
                gid='%d:data' % test.id)

        ax.plot(test.temporal['u_p'],
                test.R * test.temporal['u_p'], '--',
                label='Fit',
                gid='%d:fit' % test.id)

        xi = test.temporal['u_p'][test.index_sensitivity_min]
        xf = test.temporal['u_p'][test.index_sensitivity_max]
        yi = (test.temporal['u_y'] -
              test.temporal['u_ydark'])[test.index_sensitivity_min]
        yf = (test.temporal['u_y'] -
              test.temporal['u_ydark'])[test.index_sensitivity_max]

        ax.plot((xi, xf), (yi, yf),
                label='Fit range',
                linestyle='None',
                marker='o',
                gid='%d:marker' % test.id)

        # Todo : Standard EMVA3 asks to print on graph $\mu_{y.dark}$.
        self.set_legend(ax)


class PlotUyDark(Emva1288Plot):
    '''
    Create $\mu_{y.dark}$ plot with all instances of Processing in
    self.tests.
    '''

    name = 'Mean gray value in dark'
    title = '$\mu_{y.dark}$'
    xlabel = 'exposure time [ns]'
    ylabel = '$\mu_{y.dark}$ [DN]'

    def plot(self, test):
        ax = self.ax

        if np.mean(test.temporal['texp']) == test.temporal['texp'][0]:
            ax.plot(test.temporal['texp'],
                    test.temporal['u_ydark'],
                    'o',
                    markersize=5,
                    label='Data',
                    gid='%d:data' % test.id)
        else:
            ax.plot(test.temporal['texp'],
                    test.temporal['u_ydark'],
                    label='Data',
                    gid='%d:data' % test.id)
        ax.ticklabel_format(axis='x', style='sci', scilimits=(1, 4))
        self.set_legend(ax)


class PlotPTC(Emva1288Plot):
    '''
    Create Photon Transfer plot with all instances of Processing in
    self.tests.
    '''

    name = 'Photon Transfer'
    xlabel = '$\mu_y - \mu_{y.dark}$ [DN]'
    ylabel = '$\sigma^2_y - \sigma^2_{y.dark}$ [DN$^2$]'

    def plot(self, test):
        ax = self.ax

        X = test.temporal['u_y'] - test.temporal['u_ydark']
        Y = test.temporal['s2_y'] - test.temporal['s2_ydark']
        ax.plot(X, Y,
                label='Data',
                gid='%d:data' % test.id)

        ax.plot(X, test.K * X,
                linestyle='--',
                label='Fit',
                gid='%d:fit' % test.id)

        ax.plot((X[test.index_u_ysat], ), (Y[test.index_u_ysat], ),
                marker='o',
                linestyle='None',
                label='Saturation',
                gid='%d:marker' % test.id)

        ax.plot((X[test.index_sensitivity_min],
                 X[test.index_sensitivity_max]),
                (Y[test.index_sensitivity_min],
                 Y[test.index_sensitivity_max]),
                linestyle='None',
                marker='o',
                label='Fit range',
                gid='%d:marker' % test.id)

        # Todo : Standard EMVA3 asks to print on graph
        # $\sigma^2_{y.dark}$ and K with its one-sigma statistical
        # uncertainty.
        self.set_legend(ax)


class PlotSNR(Emva1288Plot):
    '''
    Create SNR plot with all instances of Processing in self.tests.
    '''

    name = 'SNR'
    xlabel = '$\mu_{p}$ [mean number of photons/pixel]'
    ylabel = 'SNR'

    def setup_figure(self):
        super(PlotSNR, self).setup_figure()
        self.ax.loglog()
        max_ideal = []
        self.max_ideal = max_ideal

    def plot(self, test):
        ax = self.ax

        X = np.arange(test.u_p_min, test.u_p_sat,
                      (test.u_p_sat - test.u_p_min) / 100.0)

        # remove the zeros on the denominator, at saturation the temporal
        # noise is zero!
        nz = np.nonzero(test.temporal['s2_y'])
        ax.plot(test.temporal['u_p'][nz],
                (test.temporal['u_y'] - test.temporal['u_ydark'])[nz] /
                np.sqrt(test.temporal['s2_y'][nz]),
                label='Data',
                gid='%d:data' % test.id)

        ax.plot(X,
                ((test.QE / 100) * X) /
                np.sqrt((test.sigma_d) ** 2 +
                        (test.s2q / (test.K) ** 2) +
                        ((test.QE / 100) * X)),
                linestyle='--',
                label='Theoretical',
                gid='%d:fit' % test.id)

        ideal = np.sqrt(X)
        self.max_ideal.append(ideal[-1])
        ax.plot((X),
                ideal,
                linestyle='--',
                label='Ideal',
                gid='%d:fit' % test.id)

        ax.axvline(test.u_p_min,
                   linestyle='--',
                   label='$\mu_{p.min} = %.1f[p]$' % test.u_p_min,
                   gid='%d:marker' % test.id)

        ax.axvline(test.u_p_sat,
                   linestyle='--',
                   label='$\mu_{p.sat} = %.1f[p]$' % test.u_p_sat,
                   gid='%d:marker' % test.id)

        ax.plot(X,
                ((test.QE / 100) * X) /
                np.sqrt((test.sigma_d) ** 2 +
                        (test.s2q / (test.K) ** 2) +
                        ((test.QE / 100) * X) +
                        (test.DSNU1288 ** 2) +
                        (((test.PRNU1288 / 100) *
                          (test.QE / 100.) * X) ** 2)),
                linestyle='--',
                label='Total SNR',
                gid='%d:fit' % test.id)

        ax.set_ylim(1, max(self.max_ideal))
        self.set_legend(ax)


class PlotLinearity(Emva1288Plot):
    '''
    Create Linearity plot with all instances of Processing in self.tests.
    '''

    name = 'Linearity'
    xlabel = '$\mu_{p}$ [mean number of photons/pixel]'
    ylabel = '$\mu_y - \mu_{y.dark}$ [DN]'

    def plot(self, test):
        ax = self.ax

        X = test.temporal['u_p']
        Y = test.temporal['u_y'] - test.temporal['u_ydark']
        ax.plot(X, Y,
                label='Data',
                gid='%d:data' % test.id)

        ax.plot(X,
                test.linearity()['fit_slope'] *
                X + test.linearity()['fit_offset'],
                linestyle='--',
                label='Fit',
                gid='%d:fit' % test.id)

        ax.plot((X[test.index_linearity_min], X[test.index_linearity_max]),
                (Y[test.index_linearity_min], Y[test.index_linearity_max]),
                label='Fit range',
                linestyle='None',
                marker='o',
                gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotDeviationLinearity(Emva1288Plot):
    '''
    Create Deviation Linearity plot with all instances of Processing in
    self.tests.
    '''

    name = 'Deviation linearity'
    xlabel = '$\mu_{p}$ [mean number of photons/pixel]'
    ylabel = 'Linearity error LE [%]'

    def plot(self, test):
        ax = self.ax

        X = test.temporal['u_p'][test.index_linearity_min:
                                 test.index_linearity_max]
        deviation = test.linearity()['relative_deviation']
        Y = deviation[test.index_linearity_min: test.index_linearity_max]
        ax.plot(X, Y,
                label='Data',
                gid='%d:data' % test.id)

        ax.plot((X[0], X[-1]),
                (Y[0], Y[-1]),
                label='Fit range',
                linestyle='None',
                marker='o',
                gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotHorizontalSpectogramPRNU(Emva1288Plot):
    '''
    Create Horizontal spectrogram PRNU plot with all instances of
    Processing in self.tests.
    '''

    name = 'Horizontal spectrogram PRNU'
    xlabel = 'cycles [periods/pixel]'
    ylabel = 'Standard deviation and\nrelative presence of each cycle [DN]'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        # In Release 3.2, there is no subtraction of the residue.
        spectrogram = routines.FFT1288(test.spatial['avg'][0])

        ax.plot(routines.GetFrecs(spectrogram[:(np.shape(spectrogram)[0] //
                                                2)]),
                (np.sqrt(spectrogram[:(np.shape(spectrogram)[0] // 2)])),
                label='Data',
                gid='%d:data' % test.id)

        ax.axhline(np.sqrt(test.sigma_2_y_stack),
                   label='$\sigma^2_{y.stack}$',
                   linestyle='--',
                   gid='%d:marker' % test.id)

        # TODO: Standard EMVA3 asks to print on graph s_w and F.
        self.set_legend(ax)


class PlotHorizontalSpectrogramDSNU(Emva1288Plot):
    '''
    Create Horizontal spectrogram DSNU plot with all instances of
    Processing in self.tests.
    '''

    name = 'Horizontal spectrogram DSNU'
    xlabel = 'cycles [periods/pixel]'
    ylabel = 'Standard deviation and\nrelative presence of each cycle [DN]'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        spectrogram = routines.FFT1288(test.spatial['avg_dark'][0])
        ax.plot(routines.GetFrecs(spectrogram[:(np.shape(spectrogram)[0] //
                                                2)]),
                np.sqrt(spectrogram[:(np.shape(spectrogram)[0] // 2)]),
                label='Data',
                gid='%d:data' % test.id)

        ax.axhline(np.sqrt(test.sigma_2_y_stack_dark),
                   label='$\sigma^2_{y.stack.dark}$',
                   linestyle='--',
                   gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotVerticalSpectrogramPRNU(Emva1288Plot):
    '''
    Create Vertical spectrogram PRNU plot with all instances of Processing
    in self.tests.
    '''

    name = 'Vertical spectrogram PRNU'
    xlabel = 'cycles [periods/pixel]'
    ylabel = 'Standard deviation and\nrelative presence of each cycle [DN]'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        spectrogram = routines.FFT1288(test.spatial['avg'][0], rotate=True)

        ax.plot((routines.GetFrecs(spectrogram[:(np.shape(spectrogram)[0] //
                                                 2)])),
                (np.sqrt(spectrogram[:(np.shape(spectrogram)[0] // 2)])),
                label='Data',
                gid='%d:data' % test.id)

        ax.axhline(np.sqrt(test.sigma_2_y_stack),
                   label='$\sigma^2_{y.stack}$',
                   linestyle='--',
                   gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotVerticalSpectrogramDSNU(Emva1288Plot):
    '''
    Create Vertical spectrogram DSNU plot with all instances of
    Processing in self.tests.
    '''

    name = 'Vertical spectrogram DSNU'
    xlabel = 'Cycles [periods/pixel]'
    ylabel = 'Standard deviation and\nrelative presence of each cycle [DN]'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        spectrogram = routines.FFT1288(test.spatial['avg_dark'][0],
                                       rotate=True)
        ax.plot(routines.GetFrecs(spectrogram[:(np.shape(spectrogram)[0] //
                                                2)]),
                np.sqrt(spectrogram[:(np.shape(spectrogram)[0] // 2)]),
                label='Data',
                gid='%d:data' % test.id)

        ax.axhline(np.sqrt(test.sigma_2_y_stack_dark),
                   label='$\sigma^2_{y.stack.dark}$',
                   linestyle='--',
                   gid='%d:marker' % test.id)

        self.set_legend(ax)


class PlotLogarithmicHistogramDSNU(Emva1288Plot):
    '''
    Create Logarithmic histogram DSNU plot with all instances of
    Processing in self.tests.
    '''

    name = 'Logarithmic histogram DSNU'
    xlabel = 'Deviation from the mean [DN]'
    ylabel = 'Number of pixels'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        hist = test.histogram_DSNU

        ax.plot(hist['bins'], hist['values'],
                gid='%d:data' % test.id,
                label='Data')
        ax.plot(hist['bins'], hist['model'],
                '--',
                gid='%d:fit' % test.id,
                label='Model')

        self.set_legend(ax)

        ax.axis(ymin=1.0)


class PlotLogarithmicHistogramPRNU(Emva1288Plot):
    '''
    Create Logarithmic histogram PRNU plot with all instances of
    Processing in self.tests.
    '''

    name = 'Logarithmic histogram PRNU'
    xlabel = 'Deviation from the mean [%]'
    ylabel = 'Number of pixels'
    yscale = 'log'

    def plot(self, test):
        ax = self.ax
        hist = test.histogram_PRNU

        ax.plot(hist['bins'], hist['values'],
                gid='%d:data' % test.id,
                label='Data')
        ax.plot(hist['bins'], hist['model'], '--',
                gid='%d:fit' % test.id,
                label='Model')

        self.set_legend(ax)
        ax.axis(ymin=0.5)


class PlotAccumulatedLogHistogramDSNU(Emva1288Plot):
    '''
    Create Accumulated log histogram DSNU plot with all instances of
    Processing in self.tests.
    '''

    name = 'Accumulated log histogram DSNU'
    xlabel = 'Minimal deviation from the mean [DN]'
    ylabel = 'Percentage of pixels\ndeviating from the mean at least of : '
    yscale = 'log'

    def plot(self, test):
        ax = self.ax
        hist = test.histogram_DSNU_accumulated

        ax.plot(hist['bins'], hist['values'],
                gid='%d:data' % test.id,
                label='Data')

        self.set_legend(ax)


class PlotAccumulatedLogHistogramPRNU(Emva1288Plot):
    '''
    Create Accumulated log histogram PRNU plot with all instances of
    Processing in self.tests.
    '''

    name = 'Accumulated log histogram PRNU'
    xlabel = 'Minimal deviation from the mean [%]'
    ylabel = 'Percentage of pixels\ndeviating from the mean at least of : '
    yscale = 'log'

    def plot(self, test):
        ax = self.ax

        hist = test.histogram_PRNU_accumulated

        ax.plot(hist['bins'], hist['values'],
                gid='%d:data' % test.id,
                label='Data')

        self.set_legend(ax)


class ProfileBase(Emva1288Plot):
    def __init__(self, *args, **kwargs):
        Emva1288Plot.__init__(self, *args, **kwargs)
        # Dict to keep track of max and min values to adjust the
        # limit of the axis
        self.axis_limits = {'bright': {'max': [], 'min': [], 'length': []},
                            'dark': {'max': [], 'min': [], 'length': []}}

    def _get_extremes(self, mean_, min_, max_):
        min_min_i = np.argmax(mean_ - min_)
        min_min = min_[min_min_i]
        min_perc = np.abs(100. - (min_min * 100. / mean_[min_min_i]))
        min_label = 'Min ({:.1f} {:.1f}%)'.format(min_min,
                                                  min_perc)

        max_max_i = np.argmax(max_ - min_)
        max_max = max_[max_max_i]
        max_perc = np.abs(100. - (max_max * 100. / mean_[max_max_i]))
        max_label = 'Max ({:.1f} {:.1f}%)'.format(max_max,
                                                  max_perc)

        return {'min_deviation': min_min, 'min_precentage': min_perc,
                'min_label': min_label, 'max_label': max_label,
                'max_deviation': max_max, 'max_percentage': max_perc}

    def _get_image_profiles(self, image, transpose):
        if transpose:
            img = np.transpose(image)
        else:
            img = image
        profile = np.mean(img, axis=0)
        profile_min = np.min(img, axis=0)
        profile_max = np.max(img, axis=0)

        mid_i = np.shape(img)[0]
        profile_mid = img[mid_i // 2, :]
        length = np.shape(profile)[0]

#         extremes = self._get_extremes(profile, profile_min, profile_max)

        d = {'mean': profile,
             'min': profile_min,
             'max': profile_max,
             'length': length,
             'mid': profile_mid}
#         d.update(extremes)
        return d

    def get_profiles(self, bright, dark, transpose):
        b_p = self._get_image_profiles(bright, transpose)
        b_mean = np.mean(b_p['mean'])
        self.axis_limits['bright']['length'].append(b_p['length'])
        self.axis_limits['bright']['min'].append(0.9 * b_mean)
        self.axis_limits['bright']['max'].append(1.1 * b_mean)

        d_p = self._get_image_profiles(dark, transpose)
        self.axis_limits['dark']['length'].append(d_p['length'])
        self.axis_limits['dark']['min'].append(0.9 * np.mean(d_p['min']))
        self.axis_limits['dark']['max'].append(1.1 * np.mean(d_p['max']))

        return {'bright': b_p, 'dark': d_p}


class PlotHorizontalProfile(ProfileBase):
    '''
    Create Horizontal profile plot with all instances of Processing in
    self.tests.
    Profile is done with spatial images.
    '''

    name = 'Horizontal profile'

    def setup_figure(self):
        self.ax = self.figure.add_subplot(211)
        self.ax2 = self.figure.add_subplot(212)
        self.figure.suptitle(self.name)
        self.ax.set_title('PRNU')
        self.ax.set_ylabel('Vertical line [DN]')
        self.ax2.set_title('DSNU')
        self.ax2.set_xlabel('Index of the line')
        self.ax2.set_ylabel('Vertical line [DN]')

    def plot(self, test):
        ax = self.ax
        ax2 = self.ax2

        bimg = test.spatial['avg'][0]
        dimg = test.spatial['avg_dark'][0]
        profiles = self.get_profiles(bimg, dimg, False)

        x = np.arange(profiles['bright']['length'])
        ax.plot(x, profiles['bright']['mid'],
                label='Mid',
                gid='%d:marker' % test.id)
        ax.plot(x, profiles['bright']['min'],
                label='Min',
                gid='%d:marker' % test.id)
        ax.plot(x, profiles['bright']['max'],
                label='Max',
                gid='%d:marker' % test.id)
        ax.plot(x, profiles['bright']['mean'],
                label='Mean',
                gid='%d:marker' % test.id)
        ax.set_xticks([])

        x_dark = np.arange(profiles['dark']['length'])
        ax2.plot(x_dark, profiles['dark']['mid'],
                 label='Mid',
                 gid='%d:data' % test.id)
        ax2.plot(x_dark, profiles['dark']['min'],
                 label='Min',
                 gid='%d:data' % test.id)
        ax2.plot(x_dark, profiles['dark']['max'],
                 label='Max',
                 gid='%d:data' % test.id)
        ax2.plot(x_dark, profiles['dark']['mean'],
                 label='Mean',
                 gid='%d:data' % test.id)

        ax.axis(ymin=min(self.axis_limits['bright']['min']),
                ymax=max(self.axis_limits['bright']['max']),
                xmax=max(self.axis_limits['bright']['length']))
        ax2.axis(ymin=min(self.axis_limits['dark']['min']),
                 ymax=max(self.axis_limits['dark']['max']),
                 xmax=max(self.axis_limits['dark']['length']))

        self.set_legend(ax)
        self.set_legend(ax2)


class PlotVerticalProfile(ProfileBase):
    '''
    Create Vertical profile plot with all instances of Processing in
    self.tests.
    Profile is done with spatial images.
    '''

    name = 'Vertical profile'

    def setup_figure(self):
        self.ax = self.figure.add_subplot(121)
        self.ax2 = self.figure.add_subplot(122)
        self.figure.suptitle(self.name)
        self.ax.set_title('DSNU')
        self.ax.set_xlabel('Vertical line [DN]')
        self.ax.set_ylabel('Index of the line')
        self.ax2.set_title('PRNU')
        self.ax2.set_xlabel('Vertical line [DN]')

    def plot(self, test):
        ax = self.ax
        ax2 = self.ax2

        bimg = test.spatial['avg'][0] - test.spatial['avg_dark'][0]
        dimg = test.spatial['avg_dark'][0]
        profiles = self.get_profiles(bimg, dimg, True)

        y = np.arange(profiles['bright']['length'])
        ax2.plot(profiles['bright']['mid'], y,
                 label='Mid',
                 gid='%d:marker' % test.id)
        ax2.plot(profiles['bright']['min'], y,
                 label='Min',
                 gid='%d:marker' % test.id)
        ax2.plot(profiles['bright']['max'], y,
                 label='Max',
                 gid='%d:marker' % test.id)
        ax2.plot(profiles['bright']['mean'], y,
                 label='Mean',
                 gid='%d:marker' % test.id)

        ax2.set_yticks([])

        y_dark = np.arange(profiles['dark']['length'])
        ax.plot(profiles['dark']['mid'], y_dark,
                label='Mid',
                gid='%d:marker' % test.id)
        ax.plot(profiles['dark']['min'], y_dark,
                label='Min',
                gid='%d:marker' % test.id)
        ax.plot(profiles['dark']['max'], y_dark,
                label='Max',
                gid='%d:marker' % test.id)
        ax.plot(profiles['dark']['mean'], y_dark,
                label='Mean',
                gid='%d:marker' % test.id)

        ax2.axis(xmin=min(self.axis_limits['bright']['min']),
                 xmax=max(self.axis_limits['bright']['max']),
                 ymax=max(self.axis_limits['bright']['length']))
        ax.axis(xmin=min(self.axis_limits['dark']['min']),
                xmax=max(self.axis_limits['dark']['max']),
                ymax=max(self.axis_limits['dark']['length']))

        self.set_legend(ax)
        self.set_legend(ax2)


EVMA1288plots = [PlotPTC,
                 PlotSNR,
                 PlotSensitivity,
                 PlotUyDark,
                 PlotLinearity,
                 PlotDeviationLinearity,
                 PlotHorizontalSpectogramPRNU,
                 PlotHorizontalSpectrogramDSNU,
                 PlotVerticalSpectrogramPRNU,
                 PlotVerticalSpectrogramDSNU,
                 PlotLogarithmicHistogramDSNU,
                 PlotLogarithmicHistogramPRNU,
                 PlotAccumulatedLogHistogramDSNU,
                 PlotAccumulatedLogHistogramPRNU,
                 PlotHorizontalProfile,
                 PlotVerticalProfile]


class Plotting1288(object):
    def __init__(self, *args, **kwargs):
        '''
        Creates and shows all plots necessary to prepare a camera or sensor
        descriptive report compliant with EMVA Standard 1288 version 3.1.
        '''

        self._titles = kwargs.pop('titles', True)
        # Get data
        self.tests = args

    def plots_to_plot(self, *plots):
        p = []
        if not plots:
            plots = range(len(EVMA1288plots))
        for i in plots:
            if i not in range(len(EVMA1288plots)):
                print('Error ', i, 'is not valid index')
                print('Plot has to be integer in ', range(len(EVMA1288plots)))
                continue
            p.append(i)
        return p

    def plot(self, *ids, **kwargs):
        import matplotlib.pyplot as plt
        show = kwargs.pop('show', True)
        savedir = kwargs.pop('savedir', False)
        plots = self.plots_to_plot(*ids)
        for i in plots:
            figure = plt.figure(i)
            self.plot_figure(i, figure)
        if show:
            plt.show()

        if not savedir:
            return

        for i in plots:
            figure = plt.figure(i)
            fname = EVMA1288plots[i].__name__ + '.pdf'
            fname = os.path.join(savedir, fname)
            figure.savefig(fname)

    def plot_figure(self, i, figure):
        plot = EVMA1288plots[i](figure)
        for test in self.tests:
            plot.add_test(test)
        if not self._titles:
            figure.canvas.set_window_title('Fig %d' % (i + 1))
        else:
            figure.canvas.set_window_title(plot.name)



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Transform image data into data
This class takes a dictionary with image data (product of loader.LoadImageData)
and transforms it into data that can be used for the EMVA1288 computations.
It is important to note, that this is separate from LoadImageData because this
step, produces float values that are not easily transportable (db, json,
etc...) without loosing accuracy.

"""

import numpy as np
import logging


class Data1288(object):
    '''
    Take data from parsed images (descriptor file)
    and expose it as temporal and spatial dicts
    This dicts have the form appropiate for the processing
    '''

    def __init__(self,
                 data,
                 loglevel=logging.INFO):

        self.pixels = data['format']['width'] * data['format']['height']

        logging.basicConfig()
        self.log = logging.getLogger('Data')
        self.log.setLevel(loglevel)

        self.data = {}
        self.data['name'] = data.get('name', 'Unkown')
        self.data['format'] = data['format']
        self.data['temporal'] = self._get_temporal(data['temporal'])
        self.data['spatial'] = self._get_spatial(data['spatial'])

    def _get_float_keys(self, d, depth=2):
        r = {}
        for k in d.keys():
            if depth > 1:
                r[float(k)] = self._get_float_keys(d[k], depth - 1)
            else:
                r[float(k)] = d[k]
        return r

    def _get_temporal(self, data):
        '''
        Fill the temporal dict, with the stuff that we need.
        Compute the averages and variances from the sums (sum and pvar)
        '''
        temporal = {}

        bright = self._get_float_keys(data['bright'])
        dark = self._get_float_keys(data['dark'], depth=1)

        assert not set(bright.keys()) - set(dark.keys()), \
            'Dark and bright must have same exposures'

        texp = np.asarray(sorted(bright.keys()))
        temporal['texp'] = texp

        u_p = []
        u_y = []
        s2_y = []
        u_ydark = []
        s2_ydark = []

        for t in texp:
            photons = sorted(bright[t].keys())
            for p in photons:
                u_p.append(p)
                d = self._get_temporal_data(bright[t][p])
                u_y.append(d['mean'])
                s2_y.append(d['var'])

            d = self._get_temporal_data(dark[t])
            u_ydark.append(d['mean'])
            s2_ydark.append(d['var'])

        temporal['u_p'] = np.asarray(u_p)
        temporal['u_y'] = np.asarray(u_y)
        temporal['s2_y'] = np.asarray(s2_y)
        temporal['u_ydark'] = np.asarray(u_ydark)
        temporal['s2_ydark'] = np.asarray(s2_ydark)

        # In case we have only one exposure, we need arrays with the
        # same length as the up
        # we just repeat the same value over and over
        if len(temporal['texp']) == 1:
            l = len(temporal['u_p'])

            v = temporal['texp'][0]
            temporal['texp'] = np.asarray([v for _i in range(l)])

            v = temporal['u_ydark'][0]
            temporal['u_ydark'] = np.asarray([v for _i in range(l)])

            v = temporal['s2_ydark'][0]
            temporal['s2_ydark'] = np.asarray([v for _i in range(l)])

        return temporal

    def _get_temporal_data(self, d):
        '''Convert temporal image data to mean and variance

        The mean is the sum of the pixles of the two images divided by
        (2 * self.pixels)

        The variance is the pseudo variance(integer), divided by
        (4 * self.pixels)
        '''
        mean_ = d['sum'] / (2.0 * self.pixels)
        var_ = d['pvar'] / (4.0 * self.pixels)
        return {'mean': mean_, 'var': var_}

    def _get_spatial(self, data):
        '''Fill the spatial dict
        The images (sum and pvar) are preserved, they are needed for processing
        '''
        spatial = {}

        bright = self._get_float_keys(data['bright'])
        dark = self._get_float_keys(data['dark'], depth=1)

        assert not set(bright.keys()) - set(dark.keys()), \
            'Dark and bright must have same exposures'

        texp = np.asarray(sorted(bright.keys()))
        spatial['texp'] = texp

        u_p = []
        _sum = []
        _pvar = []
        _L = []
        _avg = []
        _var = []

        _sum_dark = []
        _pvar_dark = []
        _L_dark = []
        _avg_dark = []
        _var_dark = []

        for t in texp:
            photons = sorted(bright[t].keys())

            for p in photons:
                d = self._get_spatial_data(bright[t][p])

                u_p.append(p)
                _sum.append(d['sum'])
                _pvar.append(d['pvar'])
                _L.append(d['L'])
                _avg.append(d['avg'])
                _var.append(d['var'])

            d = self._get_spatial_data(dark[t])

            _sum_dark.append(d['sum'])
            _pvar_dark.append(d['pvar'])
            _L_dark.append(d['L'])
            _avg_dark.append(d['avg'])
            _var_dark.append(d['var'])

        spatial['u_p'] = np.asarray(u_p)
        spatial['sum'] = np.asarray(_sum)
        spatial['pvar'] = np.asarray(_pvar)
        spatial['L'] = np.asarray(_L)
        spatial['avg'] = np.asarray(_avg)
        spatial['var'] = np.asarray(_var)

        spatial['sum_dark'] = np.asarray(_sum_dark)
        spatial['pvar_dark'] = np.asarray(_pvar_dark)
        spatial['L_dark'] = np.asarray(_L_dark)
        spatial['avg_dark'] = np.asarray(_avg_dark)
        spatial['var_dark'] = np.asarray(_var_dark)

        return spatial

    def _get_spatial_data(self, d):
        '''Add the mean and variance to the spatial image data

        The mean is the sum of the images divided by L

        The variance is the pseudovariance divided by
        (L^2 * (L-1))
        '''
        sum_ = np.asarray(d['sum'], dtype=np.int64)
        pvar_ = np.asarray(d['pvar'], dtype=np.int64)
        L = int(d['L'])
        avg_ = sum_ / (1.0 * L)
        var_ = pvar_ / (1.0 * np.square(L) * (L - 1))

        return {'sum': sum_,
                'pvar': pvar_,
                'L': L,
                'avg': avg_,
                'var': var_}

    def __getitem__(self, key):
        '''Adds the possibility to pass the object directly to Results1288
        Instead of using
        d = Data1288(xxx)
        r = Results1288(d.data)
        It is possible to use
        d = Data1288(xxx)
        r = Results1288(d)
        '''
        return self.data[key]



from setuptools import setup

from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

setup(name='emva1288',
      packages=['emva1288', 'emva1288.report'],
      version='0.2.0',
      description='EMVA1288 reference implementation',
      long_description=long_description,
      author='Federico Ariza',
      author_email='ariza.federico@gmail.com',
      url='https://github.com/EMVA1288/emva1288',
      keywords=['sensors', 'cameras'],
      classifiers=['Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.4'],
      install_requires=['matplotlib',
                        'numpy',
                        'pillow',
                        'lxml',
                        'scipy',
                        'jinja2'],
      package_data={'emva1288': ['report/files/*', 'report/templates/*',
                                 'examples/*']}
      )
